"""
    This file is part of Accountant.

    Accountant is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Accountant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with Accountant.  If not, see <http://www.gnu.org/licenses/>.
"""
from flask.ext.restplus import Resource, fields, marshal_with_field

from accountant import db

from .. import api

from ..models.accounts import Account
from ..models.operations import Operation

from ..fields import Object

from .models import (account_model, solds_model, balance_model,
                     category_model, ohlc_model)
from .parsers import range_parser
from .users import requires_auth


ns = api.namespace('account', description='Account management')


@ns.route('/')
@api.doc(
    security='apikey',
    responses={
        401: 'Unauthorized'
    })
class AccountListResource(Resource):
    @requires_auth
    @api.response(200, 'OK', [account_model])
    @marshal_with_field(fields.List(Object(account_model)))
    def get(self):
        """
        Returns accounts with their balances.
        """
        return Account.query().all(), 200

    @requires_auth
    @api.expect(account_model)
    @api.response(201, 'Account created', account_model)
    @api.response(406, 'Invalid account data')
    @marshal_with_field(Object(account_model))
    def post(self):
        """
        Create a new account.
        """
        data = api.payload

        # A new account MUST NOT have an id;
        if 'id' in data and data['id']:
            api.abort(
                406,
                error_message='Id must not be provided on creation.'
            )

        # Instantiate account with data.
        account = Account(**data)

        # Add new account in session.
        db.session.add(account)

        # Flush session to have id in account.
        db.session.flush()

        # Return account.
        return account, 201


@ns.route('/<int:id>')
@api.doc(
    security='apikey',
    params={
        'id': 'Id of the account to manage'
    },
    responses={
        401: 'Unauthorized',
        404: 'Account not found'
    })
class AccountResource(Resource):
    @requires_auth
    @api.response(200, 'OK', account_model)
    @marshal_with_field(Object(account_model))
    def get(self, id):
        """
        Get an account.
        """
        account = Account.query().get(id)

        if not account:
            api.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account, 200

    @requires_auth
    @api.expect(account_model)
    @api.response(200, 'OK', account_model)
    @api.response(406, 'Invalid account data')
    @marshal_with_field(Object(account_model))
    def post(self, id):
        """
        Update an account.
        """
        data = api.payload

        # Check ID consistency.
        if 'id' in data and data['id'] and data['id'] != id:
            api.abort(
                406,
                error_message='Id must not be provided or changed on update.'
            )

        # Need to get the object to update it.
        account = Account.query().get(id)

        if not account:
            api.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # SQLAlchemy objects ignore __dict__.update() with merge.
        for k, v in data.items():
            setattr(account, k, v)

        db.session.merge(account)

        # Return account.
        return account, 200

    @requires_auth
    @api.response(204, 'Account deleted', account_model)
    @marshal_with_field(Object(account_model))
    def delete(self, id):
        """
        Delete an account.
        """

        # Need to get the object to update it.
        account = Account.query().get(id)

        if not account:
            api.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        db.session.delete(account)

        return None, 204


@ns.route('/<int:id>/solds')
class SoldsResource(Resource):
    @requires_auth
    @api.doc(
        security='apikey',
        responses={
            200: ('OK', solds_model),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @marshal_with_field(Object(solds_model))
    def get(self, id):
        """
        Get solds for a specific account and date range.
        """
        account = Account.query().get(id)

        if not account:
            api.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account.solds(), 200


@ns.route('/<int:id>/balance')
class BalanceResource(Resource):
    @requires_auth
    @api.doc(
        security='apikey',
        responses={
            200: ('OK', balance_model),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @api.expect(range_parser)
    @marshal_with_field(Object(balance_model))
    def get(self, id):
        """
        Get account balance for a specific date range.
        """
        account = Account.query().get(id)

        if not account:
            api.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        data = range_parser.parse_args()

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account.balance(**data), 200


@ns.route("/<int:id>/category")
class CategoryResource(Resource):
    @requires_auth
    @api.doc(
        security='apikey',
        responses={
            200: ('OK', [category_model]),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @api.expect(range_parser)
    @marshal_with_field(fields.List(Object(category_model)))
    def get(self, id):
        """
        Get account category balances for a specific date range.
        """
        data = range_parser.parse_args()

        return Operation.get_categories_for_range(id, **data).all()


@ns.route('/<int:id>/ohlc')
class OHLCResource(Resource):
    @requires_auth
    @api.doc(
        security='apikey',
        responses={
            200: ('OK', [ohlc_model]),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @api.expect(range_parser)
    @marshal_with_field(fields.List(Object(ohlc_model)))
    def get(self, id):
        """
        Get OHLC data for a specific date range and account.
        """
        data = range_parser.parse_args()

        return Operation.get_ohlc_per_day_for_range(id, **data).all()
